/*
 * Name:    mp_util.c
 *
 * Purpose: This file contains the code that initializes the underlying
 *          MX library.  It also includes other miscellaneous functions
 *          that do not fit elsewhere.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2002, 2004, 2006 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include "Python.h"

#include <signal.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_version.h"
#include "mx_handle.h"
#include "mx_plot.h"
#include "mp.h"

MX_EXPORT PyObject *
MpPrivate_init_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] = "MpPrivate_init_cmd()";

	mx_status_type mx_status;

	/* Initialize the runtime environment for MX. */

	mx_status = mx_initialize_runtime();

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	MX_DEBUG( 2,("%s invoked", fname));

	/* Initialize the drivers for the various MX record types. */

	mx_status = mx_initialize_drivers();

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Create a global handle table for the MX records.  Configure the
	 * table to be incremented in blocks of 100 handles.
	 */

	mx_status = mx_create_handle_table(
					&MpPrivate_global_record_handle_table,
					100, 1 );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Create a global handle table for MX_NETWORK_FIELD structures.
	 * Configure the table to be incremented in blocks of 100 handles.
	 */

	mx_status = mx_create_handle_table(
				&MpPrivate_global_network_field_handle_table,
				100, 1 );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_get_debug_level_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	int debug_level;

	debug_level = mx_get_debug_level();

	result = Py_BuildValue( "i", debug_level );

	return result;
}

MX_EXPORT PyObject *
MpPrivate_set_debug_level_cmd( PyObject *self, PyObject *args )
{
	int debug_level, parse_status;

	parse_status = PyArg_ParseTuple( args, "i", &debug_level );

	if ( parse_status == 0 )
		return NULL;

	mx_set_debug_level( debug_level );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_plotting_is_enabled_cmd( PyObject *self, PyObject *args )
{
	PyObject *result;
	MX_RECORD *record;
	long handle;
	int plot_enable_value, parse_status;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "l", &handle );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	plot_enable_value = mx_plotting_is_enabled( record );

	result = Py_BuildValue( "i", plot_enable_value );

	return result;
}

MX_EXPORT PyObject *
MpPrivate_set_plot_enable_cmd( PyObject *self, PyObject *args )
{
	MX_RECORD *record;
	long handle;
	int plot_enable_value, parse_status;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "li",
					&handle, &plot_enable_value );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_set_plot_enable( record, plot_enable_value );

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_set_program_name_cmd( PyObject *self, PyObject *args )
{
	MX_RECORD *record_list;
	long handle;
	char *program_name;
	int parse_status;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "ls", &handle, &program_name );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( handle, &record_list );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	mx_status = mx_set_program_name( record_list, program_name );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

/* MpPrivateErr_SetString() is a varargs version of PyErr_SetString. */

MX_EXPORT void
MpPrivateErr_SetString( PyObject *exception_object, char *format, ... )
{
	va_list args;
	char buffer[2500];

	va_start(args, format);
	vsnprintf(buffer, sizeof(buffer), format, args);
	va_end(args);

	PyErr_SetString( exception_object, buffer );

	return;
}

/* These are the exception objects for MP. */

MX_EXPORT PyObject *MpPrivateExc_Error = NULL;

MX_EXPORT PyObject *MpPrivateExc_Success = NULL;
MX_EXPORT PyObject *MpPrivateExc_Null_Argument = NULL;
MX_EXPORT PyObject *MpPrivateExc_Illegal_Argument = NULL;
MX_EXPORT PyObject *MpPrivateExc_Corrupt_Data_Structure = NULL;
MX_EXPORT PyObject *MpPrivateExc_Unparseable_String = NULL;
MX_EXPORT PyObject *MpPrivateExc_End_Of_Data = NULL;
MX_EXPORT PyObject *MpPrivateExc_Unexpected_End_Of_Data = NULL;
MX_EXPORT PyObject *MpPrivateExc_Not_Found = NULL;
MX_EXPORT PyObject *MpPrivateExc_Type_Mismatch = NULL;
MX_EXPORT PyObject *MpPrivateExc_Not_Yet_Implemented = NULL;
MX_EXPORT PyObject *MpPrivateExc_Unsupported = NULL;
MX_EXPORT PyObject *MpPrivateExc_Out_Of_Memory = NULL;
MX_EXPORT PyObject *MpPrivateExc_Would_Exceed_Limit = NULL;
MX_EXPORT PyObject *MpPrivateExc_Limit_Was_Exceeded = NULL;
MX_EXPORT PyObject *MpPrivateExc_Interface_IO_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Device_IO_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_File_IO_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Terminal_IO_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_IPC_IO_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_IPC_Connection_Lost = NULL;
MX_EXPORT PyObject *MpPrivateExc_Network_IO_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Network_Connection_Lost = NULL;
MX_EXPORT PyObject *MpPrivateExc_Not_Ready = NULL;
MX_EXPORT PyObject *MpPrivateExc_Interrupted = NULL;
MX_EXPORT PyObject *MpPrivateExc_Pause_Requested = NULL;
MX_EXPORT PyObject *MpPrivateExc_Interface_Action_Failed = NULL;
MX_EXPORT PyObject *MpPrivateExc_Device_Action_Failed = NULL;
MX_EXPORT PyObject *MpPrivateExc_Function_Failed = NULL;
MX_EXPORT PyObject *MpPrivateExc_Controller_Internal_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Permission_Denied = NULL;
MX_EXPORT PyObject *MpPrivateExc_Client_Request_Denied = NULL;
MX_EXPORT PyObject *MpPrivateExc_Try_Again = NULL;
MX_EXPORT PyObject *MpPrivateExc_Timed_Out = NULL;
MX_EXPORT PyObject *MpPrivateExc_Hardware_Configuration_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Hardware_Fault = NULL;
MX_EXPORT PyObject *MpPrivateExc_Record_Disabled_Due_To_Fault = NULL;
MX_EXPORT PyObject *MpPrivateExc_Record_Disabled_By_User = NULL;
MX_EXPORT PyObject *MpPrivateExc_Initialization_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Read_Only = NULL;
MX_EXPORT PyObject *MpPrivateExc_Software_Configuration_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Operating_System_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Unknown_Error = NULL;
MX_EXPORT PyObject *MpPrivateExc_Not_Valid_For_Current_State = NULL;
MX_EXPORT PyObject *MpPrivateExc_Configuration_Conflict = NULL;
MX_EXPORT PyObject *MpPrivateExc_Not_Available = NULL;
MX_EXPORT PyObject *MpPrivateExc_Stop_Requested = NULL;
MX_EXPORT PyObject *MpPrivateExc_Bad_Handle = NULL;
MX_EXPORT PyObject *MpPrivateExc_Object_Abandoned = NULL;
MX_EXPORT PyObject *MpPrivateExc_Might_Cause_Deadlock = NULL;
MX_EXPORT PyObject *MpPrivateExc_Already_Exists = NULL;

/*---*/

static struct {
	PyObject **exception_object_address;
	char exception_name[40];
} MpPrivate_exception_table[] = {

{ &MpPrivateExc_Success,                      "Success_Error" },
{ &MpPrivateExc_Null_Argument,                "Null_Argument_Error" },
{ &MpPrivateExc_Illegal_Argument,             "Illegal_Argument_Error" },
{ &MpPrivateExc_Corrupt_Data_Structure,       "Corrupt_Data_Structure_Error" },
{ &MpPrivateExc_Unparseable_String,           "Unparseable_String_Error" },
{ &MpPrivateExc_End_Of_Data,                  "End_Of_Data_Error" },
{ &MpPrivateExc_Unexpected_End_Of_Data,       "Unexpected_End_Of_Data_Error" },
{ &MpPrivateExc_Not_Found,                    "Not_Found_Error" },
{ &MpPrivateExc_Type_Mismatch,                "Type_Mismatch_Error" },
{ &MpPrivateExc_Not_Yet_Implemented,          "Not_Yet_Implemented_Error" },
{ &MpPrivateExc_Unsupported,                  "Unsupported_Error" },
{ &MpPrivateExc_Out_Of_Memory,                "Out_Of_Memory_Error" },
{ &MpPrivateExc_Would_Exceed_Limit,           "Would_Exceed_Limit_Error" },
{ &MpPrivateExc_Limit_Was_Exceeded,           "Limit_Was_Exceeded_Error" },
{ &MpPrivateExc_Interface_IO_Error,           "Interface_IO_Error" },
{ &MpPrivateExc_Device_IO_Error,              "Device_IO_Error" },
{ &MpPrivateExc_File_IO_Error,                "File_IO_Error" },
{ &MpPrivateExc_Terminal_IO_Error,            "Terminal_IO_Error" },
{ &MpPrivateExc_IPC_IO_Error,                 "IPC_IO_Error" },
{ &MpPrivateExc_IPC_Connection_Lost,          "IPC_Connection_Lost_Error" },
{ &MpPrivateExc_Network_IO_Error,             "Network_IO_Error" },
{ &MpPrivateExc_Network_Connection_Lost,      "Network_Connection_Lost_Error" },
{ &MpPrivateExc_Not_Ready,                    "Not_Ready_Error" },
{ &MpPrivateExc_Interrupted,                  "Interrupted_Error" },
{ &MpPrivateExc_Pause_Requested,              "Pause_Requested_Error" },
{ &MpPrivateExc_Interface_Action_Failed,      "Interface_Action_Failed_Error" },
{ &MpPrivateExc_Device_Action_Failed,         "Device_Action_Failed_Error" },
{ &MpPrivateExc_Function_Failed,              "Function_Failed_Error" },
{ &MpPrivateExc_Controller_Internal_Error,    "Controller_Internal_Error" },
{ &MpPrivateExc_Permission_Denied,            "Permission_Denied_Error" },
{ &MpPrivateExc_Client_Request_Denied,        "Client_Request_Denied_Error" },
{ &MpPrivateExc_Try_Again,                    "Try_Again_Error" },
{ &MpPrivateExc_Timed_Out,                    "Timed_Out_Error" },
{ &MpPrivateExc_Hardware_Configuration_Error, "Hardware_Configuration_Error" },
{ &MpPrivateExc_Hardware_Fault,               "Hardware_Fault_Error" },
{ &MpPrivateExc_Record_Disabled_Due_To_Fault,
					"Record_Disabled_Due_To_Fault_Error" },
{ &MpPrivateExc_Record_Disabled_By_User,      "Record_Disabled_By_User_Error" },
{ &MpPrivateExc_Initialization_Error,         "Initialization_Error" },
{ &MpPrivateExc_Read_Only,                    "Read_Only_Error" },
{ &MpPrivateExc_Software_Configuration_Error, "Software_Configuration_Error" },
{ &MpPrivateExc_Operating_System_Error,       "Operating_System_Error" },
{ &MpPrivateExc_Unknown_Error,                "Unknown_Error_Error" },
{ &MpPrivateExc_Not_Valid_For_Current_State, 
					"Not_Valid_For_Current_State_Error" },
{ &MpPrivateExc_Configuration_Conflict,       "Configuration_Conflict_Error" },
{ &MpPrivateExc_Not_Available,                "Not_Available_Error" },
{ &MpPrivateExc_Stop_Requested,               "Stop_Requested_Error" },
{ &MpPrivateExc_Bad_Handle,                   "Bad_Handle_Error" },
{ &MpPrivateExc_Object_Abandoned,             "Object_Abandoned_Error" },
{ &MpPrivateExc_Might_Cause_Deadlock,         "Might_Cause_Deadlock_Error" },
{ &MpPrivateExc_Already_Exists,               "Already_Exists_Error" },

};

static int MpPrivate_num_exceptions = sizeof( MpPrivate_exception_table )
					/ sizeof( MpPrivate_exception_table[0]);

/*---*/

/* MpPrivateErr_SetFromMxStatus() translates between MX errors
 * and Python exceptions.
 */

MX_EXPORT PyObject *
MpPrivateErr_SetFromMxStatus( PyObject *error_object, mx_status_type mx_status )
{
	int offset;

	if ( error_object == (PyObject *) NULL ) {

		offset = mx_status.code - MXE_SUCCESS;

		if ( offset < 0 ) {
			mx_warning(
		"mx_status.code %ld has a smaller value than MXE_SUCCESS (%d)."
		"  This should not be able to happen.",
				mx_status.code, MXE_SUCCESS );

			error_object = MpPrivateExc_Error;
		} else
		if ( offset > MpPrivate_num_exceptions ) {

			error_object = MpPrivateExc_Error;
		} else {
			error_object = 
		  *(MpPrivate_exception_table[offset].exception_object_address);
		}

		if ( error_object == NULL ) {
			mx_warning(
			"Mp error object for MX error code %ld is NULL."
			"  This should not be able to happen.",
				mx_status.code );
		}
	}

	PyErr_SetString( error_object, mx_status.message );

	return NULL;
}

MX_EXPORT PyObject *
MpPrivate_init_exceptions_cmd( PyObject *self, PyObject *args )
{
	PyObject *mp_module_object, *mp_module_dictionary;
	PyObject *new_exception;
	char *exception_name;
	char long_exception_name[80];
	int i, parse_status;

	/* Find the Mp module and its dictionary. */

	parse_status = PyArg_ParseTuple( args, "O", &mp_module_object );

	if ( parse_status == 0 )
		return NULL;

	mp_module_dictionary = PyModule_GetDict( mp_module_object );

	if ( mp_module_dictionary == NULL )
		return NULL;

	/* Create an Mp.Error exception class which will be the superclass
	 * for all the other exceptions.
	 */

	MpPrivateExc_Error = PyErr_NewException( "Mp.Error", NULL, NULL );

	if ( MpPrivateExc_Error == NULL )
		return NULL;

	PyDict_SetItemString( mp_module_dictionary,
				"Error", MpPrivateExc_Error );

	/* Now create all the rest of the exceptions. */

	for ( i = 0; i < MpPrivate_num_exceptions; i++ ) {
		exception_name = MpPrivate_exception_table[i].exception_name;

		snprintf( long_exception_name, sizeof(long_exception_name),
				"Mp.%s", exception_name );

		new_exception = PyErr_NewException( long_exception_name,
					MpPrivateExc_Error, NULL );

		if ( new_exception == NULL )
			return NULL;

		*(MpPrivate_exception_table[i].exception_object_address)
			= new_exception;

		PyDict_SetItemString( mp_module_dictionary,
					exception_name, new_exception );
	}

	return Py_None;
}

